/* Copyright 2007-2013, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* Syscon memory map:
 * 0x00: reads 0x7250
 * 0x02: Control register:
     bits 3-0:   FPGA revision (RO)
     bits 5-4:   Mode2, Mode1
     bits 7-6:   Scratch register
     bit 8:     Enable DB9 console
 * 0x04: external bus config register
 * 0x06: watchdog feed register
 * 0x08: free running 1MHz counter LSB
 * 0x0a: free running 1MHz counter MSB
 * 0x0c: RNG LSB
 * 0x0e: RNG MSB
 * 0x28: TAG memory access
 * 0x2a: Custom load ID register (reads 0 on standard load)
 * 0x2c: IRQ register (smc_irq line indicates one of the following is active)
 *   bit 0: XUART IRQ
 *   bit 1: CAN IRQ
 *   bit 2: CAN2 IRQ
 *   bit 3: PC/104 IRQ 5
 *   bit 4: PC/104 IRQ 6
 *   bit 5: PC/104 IRQ 7
 *   bits 6-13: reserved
 * 0x2e: IRQ mask register.  Setting a bit in this register disables the IRQ
 *       for the corresponding bit in the IRQ register.
 * 0x34: ADC pin control
 * 0x36: evgpio core #1
 * 0x38: evgpio mask core #1
 * 0x3a: evgpio core #2
 * 0x3c: evgpio mask core #2
 * 0x3e: LCD PWM 
 *   bits 15-4: reserved
 *   bits 3-0: duty cycle (0x7 = 100%)
 *
 * EVGPIO #1
 *  24-0: ISA misc (see below for misc pin mapping)
 *  40-25: ISA data bus 15-0
 *  60-41: ISA address 19-0
 *  61: ISA data dir (controls U33 245 buffer ISA databus direction)
 *  62: ISA data oe (controls U33 245 buffer output enable)
 *  63: ISA memory enable (when 1, ISA bus uses MEMR#/MEMW# strobes)
 *  88-64: ISA misc oe
 *  104-89: ISA data bus oe
 *  124-105: ISA address oe
 *  127: ISA 8-bit enable
 *
 * EVGPIO #2
 *  11-0: LCD 11-0
 *  21-12: DIO 9-0
 *  37-22: ISA_CN_D 15-0
 *  38: DB9 RTS (output only)
 *  39: DB9 DTR (output only)
 *  40: COM2 RTS (output only)
 *  41: COM3 RTS (output only)
 *  42: DB9 CTS (input only)
 *  43: DB9 DSR (input only)
 *  44: DB9 DCD (input only)
 *  45: COM2 CTS (input only)
 *  46: COM3 CTS (input only)
 *  47: JP1 (input only)
 *  48: JP2# (input only)
 *  49: M0 spare (input)
 *  50: M0 spare (output)
 *  51-63: reserved
 *  75-64: LCD oe 
 *  85-76: DIO oe
 *  101-86: ISA_CN_D oe
 *  114: M0 spare oe
 *
 * ISA misc bus breakdown:
 *  0 - isa_ior
 *  1 - isa_iow
 *  2 - isa_memw
 *  3 - isa_memr
 *  4 - isa_reset
 *  5 - isa_14_3_mhz
 *  6 - isa_endx
 *  7 - isa_aen
 *  8 - isa_refresh
 *  9 - isa_iochk
 *  10 - isa_iordy
 *  11 - isa_drq3
 *  12 - isa_drq2
 *  13 - isa_dack3
 *  14 - isa_irq5_3v
 *  15 - isa_irq6_3v
 *  16 - isa_irq7_3v
 *  17 - isa_irq9
 *  18 - isa_dack1
 *  19 - isa_drq1
 *  20 - isa_irq3
 *  21 - isa_bclk
 *  22 - isa_dack2
 *  23 - isa_tc
 *  24 - isa_bale
 */
 
/*
 *  0 - Initial revision
 *  1 - Fix ISA databus mixup
 *  2 - EVGPIO IRQ fix
 *  3 - Pullups on LCD/DIO/PC104 pins
 *      Console enable on DB9 via JP2
 *      Split IO/MEM PC104 region
 *      1Ghz/355Mhz clocking on PXA168
 *      soft JP #7 now keeps PXA168s at 800Mhz 
 *      IO bus cycles now drive upper address lines to 0
 *      Inverted isa_16bit_en bit so that the poweron default is enabled
 *  4 - Mask 14mhz clk into EVGPIO core
 *      Make IO PC104 bus cycles the default, evgpio bit must be used to enable MEMR/W
 *      Fix red/green LED reg bit positions to be compatible with 47xx
 *  5 - Fix ISA 8-bit cycles at the 0x81xx_xxxx chip-select
 *  6 - Add TS's special 16-bit ISA pin mapping on the PC104 64pin connector and
 *      make it the default
 *  7 - Workaround for TS-ADC16 ISA bus cycles (extra hold time on writes)
 *  8 - Fix blockram arbiter for XUARTs
 *  9 - Make XUART#4 RX input from the DB9
 *
 */

module syscon(
  wb_clk_i,
  wb_rst_i,

  wb_cyc_i,
  wb_stb_i,
  wb_we_i,
  wb_adr_i,
  wb_dat_i,
  wb_sel_i,
  wb_dat_o,
  wb_ack_o,

  dio_i,
  dio_oe_o,
  dio_o,
  usec_i,
  mode1_i,
  mode2_i,

  red_led_o,
  green_led_o,
  reboot_o,
  irqs_i,
  irq_o,
  sd_busy_i,
  dma_idle_i,
  adc_sda_o,
  adc_sda_i,
  adc_scl_o,
  adc_scl_i,
  adc_a_o,
  lcd_o,
  lcd_oe_o,
  lcd_i,
  isa_add_o,
  isa_add_oe_o,
  isa_add_i,
  isa_dat_o,
  isa_dat_oe_o,
  isa_dat_i,
  isa_misc_o,
  isa_misc_oe_o,
  isa_misc_i,
  isa_cn_d_o,
  isa_cn_d_oe_o,
  isa_cn_d_i,
  isa_dir_o,
  isa_data_en_o,
  isa_memory_en_o,
  isa_16_en_o,

  db9_cts_i,
  db9_dsr_i,
  db9_dcd_i,
  com2_cts_i,
  com3_cts_i,
  db9_rts_o,
  db9_dtr_o,
  com2_rts_o,
  com3_rts_o,
  en_db9con_o,

  m0_spare_i,
  m0_spare_o,
  m0_spare_oe_o,

  sd_led_o,
  jp1_i,
  jp2n_i,
  isa_ts_special16_disable_o

);

input wb_clk_i, wb_rst_i, wb_cyc_i, wb_stb_i, wb_we_i;
input [31:0] wb_adr_i;
input [15:0] wb_dat_i;
input [1:0] wb_sel_i;
output [15:0] wb_dat_o;
output wb_ack_o;

input [59:0] dio_i;
output [59:0] dio_oe_o;
output [59:0] dio_o;

input [31:0] usec_i;
input mode1_i;
input mode2_i;

output red_led_o;
output green_led_o;
output reboot_o;
input [15:0] irqs_i;
output irq_o;
input dma_idle_i, sd_busy_i;
output [1:0] sd_led_o;

input adc_sda_i, adc_scl_i;
output adc_sda_o, adc_scl_o;
output [2:0] adc_a_o;

output [11:0] lcd_o, lcd_oe_o;
input [11:0] lcd_i;
output [15:0] isa_dat_o, isa_dat_oe_o;
input [15:0] isa_dat_i;
output [19:0] isa_add_o, isa_add_oe_o;
input [19:0] isa_add_i;
output isa_dir_o, isa_data_en_o, isa_memory_en_o, isa_16_en_o;
output isa_ts_special16_disable_o;

/* 
 * 0 - isa_ior
 * 1 - isa_iow
 * 2 - isa_memw
 * 3 - isa_memr
 * 4 - isa_reset
 * 5 - isa_14_3_mhz
 * 6 - isa_endx
 * 7 - isa_aen
 * 8 - isa_refresh
 * 9 - isa_iochk
 * 10 - isa_iordy
 * 11 - isa_drq3
 * 12 - isa_drq2
 * 13 - isa_dack3
 * 14 - isa_irq5_3v
 * 15 - isa_irq6_3v
 * 16 - isa_irq7_3v
 * 17 - isa_irq9
 * 18 - isa_dack1
 * 19 - isa_drq1
 * 20 - isa_irq3
 * 21 - isa_bclk
 * 22 - isa_dack2
 * 23 - isa_tc
 * 24 - isa_bale
 */
output [24:0] isa_misc_o, isa_misc_oe_o;
input [24:0] isa_misc_i;
output [15:0] isa_cn_d_o, isa_cn_d_oe_o;
input [15:0] isa_cn_d_i;

input db9_cts_i, db9_dsr_i, db9_dcd_i, com2_cts_i, com3_cts_i;
output db9_rts_o, db9_dtr_o, com2_rts_o, com3_rts_o, en_db9con_o;

input m0_spare_i, jp1_i, jp2n_i;
output m0_spare_o, m0_spare_oe_o;

parameter wdog_default = 2'd3;

localparam [15:0] model = 16'h7250;
localparam [3:0] revision = 4'h9;
localparam [15:0] custom = 16'h0000;

reg tag_csn, tag_clk, tag_si;
wire tag_so;
reg tag_so_q;
tagmem tagmemcore(
  .CLK(tag_clk),
  .SI(tag_si),
  .SO(tag_so),
  .CS(tag_csn)
);

always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) tag_so_q <= 1'b0;
  else tag_so_q <= tag_so;
end

wire random_clk;
OSCE #(.NOM_FREQ("3.1")) rndosc (random_clk);
reg [31:0] random;
reg [15:0] rndcnt;
reg [1:0] wdogctl;
reg [9:0] wdogcnt;
reg feed_en, reboot;
wire sed_err;
wire feed_bsy, feed_rdreq;
wire [15:0] wdog_dat;
reg [4:0] blinkcnt;
assign reboot_o = reboot;
always @(posedge random_clk or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    rndcnt <= 16'd0;
    random <= 32'd0;
    wdogctl <= wdog_default;   
    wdogcnt <= 10'd0;
    blinkcnt <= 5'd0;
  end else begin
    rndcnt <= rndcnt + 1'b1;
    if (rndcnt == 16'd0) begin
      random[31:0] <= {random[30:0], wb_clk_i};
      wdogcnt <= wdogcnt + 1'b1;
      blinkcnt <= blinkcnt + 1'b1;
    end
    if (feed_rdreq) begin
      wdogctl <= wdog_dat;
      wdogcnt <= 10'd0;
    end
  end
end
reg [1:0] sd_led;
always @(*) begin
  case (wdogctl)
  2'd0: reboot = wdogcnt[4]; /* approx .338 seconds */
  2'd1: reboot = wdogcnt[7]; /* approx 2.706 seconds */
  2'd2: reboot = wdogcnt[9]; /* approx 10.824 seconds */
  2'd3: reboot = 1'b0;
  endcase

  sd_led = 2'd0;
  if (sd_led_reg[0]) begin
    if (sd_led_reg[1]) sd_led[0] = blinkcnt[4];
    else sd_led[0] = 1'b1;
  end
  if (sd_led_reg[2]) begin
    if (sd_led_reg[3]) sd_led[1] = blinkcnt[4];
    else sd_led[1] = 1'b1;
  end
end

resync_fifo watchdogfifo(
  .rst_i(wb_rst_i),
  .wrclk_i(wb_clk_i),
  .wrreq_i(feed_en),
  .data_i(wb_dat_i),
  .full_wrclk_o(feed_bsy),
  .rdclk_i(random_clk),
  .rdreq_i(feed_rdreq),
  .q_o(wdog_dat),
  .full_rdclk_o(feed_rdreq)
);

reg [59:0] dio_oe;
reg [59:0] dio;
reg [1:0] scratch;

reg red_led, green_led;
reg [15:0] irq_mask;

assign red_led_o = red_led;
assign green_led_o = green_led;

assign irq_o = (({evgpio2_irq, evgpio_irq, irqs_i[13:0]} & ~irq_mask) != 16'h0000);


reg [15:0] tagmem0, tagmem1;
reg [2:0] adc_a;
reg [3:0] pwm, pwm_cnt, sd_led_reg;
reg adc_scl, adc_sda, pwm_out;
assign adc_a_o = adc_a;
assign adc_scl_o = adc_scl;
assign adc_sda_o = adc_sda;
assign sd_led_o = sd_led;
reg en_db9con;
assign en_db9con_o = en_db9con ^ jp1_i;
reg [4:0] hi, lo;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    pwm <= 4'd0;
    pwm_cnt <= 4'd0;
    pwm_out <= 1'b0;
    adc_scl <= 1'b1;
    adc_sda <= 1'b1;
    adc_a <= 3'd0;
    tag_clk <= 1'b0;
    tag_si <= 1'b0;
    tag_csn <= 1'b0;
    en_db9con <= 1'b0;
    scratch <= 2'b00;
    red_led <= 1'b1;
    green_led <= 1'b1;
    irq_mask <= 16'h0000;
    tagmem0 <= 16'h0;
    tagmem1 <= 16'h0;
    sd_led_reg <= 4'd0;
    hi <= 5'd0;
    lo <= 5'd0;
  end else begin
    if (wb_cyc_i && wb_stb_i && wb_we_i && !wb_ack_q) case (wb_adr_i[5:0])
      6'h0: sd_led_reg <= wb_dat_i[3:0];
      6'h2: {en_db9con, scratch} <= wb_dat_i[8:6];
      6'h10: begin
        if (wb_sel_i[0]) lo[4:0] <= wb_dat_i[6:2];
        if (wb_sel_i[1]) hi[4:0] <= wb_dat_i[14:10];
      end
      6'h12: {red_led, green_led} <= wb_dat_i[12:11];
      6'h28: {tag_clk, tag_si, tag_csn} <= wb_dat_i[3:1];
      6'h2e: irq_mask <= wb_dat_i;
      6'h30: tagmem0 <= wb_dat_i;
      6'h32: tagmem1 <= wb_dat_i;
      6'h34: {adc_a, adc_scl, adc_sda} <= wb_dat_i[6:2];
      6'h3e: pwm <= wb_dat_i[3:0];
    endcase
    pwm_cnt <= pwm_cnt + 1'b1;
    if (pwm_cnt >= pwm) pwm_out <= 1'b0; else pwm_out <= 1'b1;
  end
end

reg evgpio_en, evgpiomask_en;
reg evgpio2_en, evgpio2mask_en;
wire [15:0] evgpio_dat, evgpiomask_dat;
wire [15:0] evgpio2_dat, evgpio2mask_dat;
wire evgpio_ack, evgpio_irq, evgpiomask_ack;
wire evgpio2_ack, evgpio2_irq, evgpio2mask_ack;
wire [127:0] evgpio_o, evgpiomask_o;
wire [127:0] evgpio2_o, evgpio2mask_o;
assign isa_misc_o[24:0] = evgpio_o[24:0];
assign isa_dat_o[15:0] = evgpio_o[40:25];
assign isa_add_o[19:0] = evgpio_o[60:41];
assign isa_dir_o = evgpio_o[61];
assign isa_data_en_o = evgpio_o[62];
assign isa_memory_en_o = evgpio_o[63];
assign isa_misc_oe_o[24:0] = evgpio_o[88:64];
assign isa_dat_oe_o[15:0] = evgpio_o[104:89];
assign isa_add_oe_o[19:0] = evgpio_o[124:105];
assign isa_ts_special16_disable_o = evgpio_o[126];
assign isa_16_en_o = !evgpio_o[127];
wire [127:0] evgpio_i, evgpio2_i;
assign evgpio_i[24:0] = isa_misc_i;
assign evgpio_i[40:25] = isa_dat_i;
assign evgpio_i[60:41] = isa_add_i;

assign lcd_o[11:4] = evgpio2_o[11:4];
assign lcd_o[3] = evgpio2_o[3] ^ pwm_out;
assign lcd_o[2:0] = evgpio2_o[2:0];
assign dio_o[9:0] = evgpio2_o[21:12];
assign isa_cn_d_o[15:0] = evgpio2_o[37:22];
assign db9_rts_o = evgpio2_o[38];
assign db9_dtr_o = evgpio2_o[39]; 
assign com2_rts_o = evgpio2_o[40];
assign com3_rts_o = evgpio2_o[41];
assign m0_spare_o = evgpio2_o[50];
assign lcd_oe_o[11:0] = evgpio2_o[75:64];
assign dio_oe_o[9:0] = evgpio2_o[85:76];
assign isa_cn_d_oe_o = evgpio2_o[101:86];
assign m0_spare_oe_o = evgpio2_o[114];
assign evgpio2_i[11:0] = lcd_i[11:0];
assign evgpio2_i[21:12] = dio_i[9:0];
assign evgpio2_i[37:22] = isa_cn_d_i[15:0];
assign evgpio2_i[42] = db9_cts_i;
assign evgpio2_i[43] = db9_dsr_i;
assign evgpio2_i[44] = db9_dcd_i;
assign evgpio2_i[45] = com2_cts_i;
assign evgpio2_i[46] = com3_cts_i;
assign evgpio2_i[47] = jp1_i;
assign evgpio2_i[48] = jp2n_i;
assign evgpio2_i[49] = m0_spare_i;

wb_evgpio evgpiocore(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && evgpio_en && !wb_ack_q),
  .wb_stb_i(wb_stb_i && evgpio_en && !wb_ack_q),
  .wb_we_i(wb_we_i),
  .wb_dat_o(evgpio_dat),
  .wb_dat_i(wb_dat_i),
  .wb_ack_o(evgpio_ack),

  .gpios_i(evgpio_i & (~evgpiomask_o)),
  .gpios_o(evgpio_o),
  .irq_o(evgpio_irq)
);

wb_evgpio evgpiomaskcore(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && evgpiomask_en && !wb_ack_q),
  .wb_stb_i(wb_stb_i && evgpiomask_en && !wb_ack_q),
  .wb_we_i(wb_we_i),
  .wb_dat_o(evgpiomask_dat),
  .wb_dat_i(wb_dat_i),
  .wb_ack_o(evgpiomask_ack),

  .gpios_o(evgpiomask_o)
);

wb_evgpio evgpio2core(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && evgpio2_en && !wb_ack_q),
  .wb_stb_i(wb_stb_i && evgpio2_en && !wb_ack_q),
  .wb_we_i(wb_we_i),
  .wb_dat_o(evgpio2_dat),
  .wb_dat_i(wb_dat_i),
  .wb_ack_o(evgpio2_ack),

  .gpios_i(evgpio2_i & (~evgpio2mask_o)),
  .gpios_o(evgpio2_o),
  .irq_o(evgpio2_irq)
);

wb_evgpio evgpio2maskcore(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && evgpio2mask_en && !wb_ack_q),
  .wb_stb_i(wb_stb_i && evgpio2mask_en && !wb_ack_q),
  .wb_we_i(wb_we_i),
  .wb_dat_o(evgpio2mask_dat),
  .wb_dat_i(wb_dat_i),
  .wb_ack_o(evgpio2mask_ack),

  .gpios_o(evgpio2mask_o)
);


reg [15:0] wb_dat, wb_dat_q;
assign wb_dat_o = wb_dat_q;
reg wb_ack, wb_ack_q;
assign wb_ack_o = wb_ack_q;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    wb_dat_q <= 16'd0;
    wb_ack_q <= 1'b0;
  end else begin
    wb_ack_q <= wb_ack;
    wb_dat_q <= wb_dat;
  end
end

always @(*) begin
  wb_ack = wb_cyc_i && wb_stb_i;
  wb_dat = 16'hxxxx;
  evgpiomask_en = 1'b0;
  evgpio_en = 1'b0;
  evgpio2mask_en = 1'b0;
  evgpio2_en = 1'b0;
  feed_en = 1'b0;

  case (wb_adr_i[5:0]) 
    6'h0: wb_dat = model[15:0];
    6'h2: wb_dat = {{7{1'b0}}, en_db9con, scratch, jp2n_i, jp1_i, revision};
    6'h6: begin
      feed_en = wb_cyc_i & wb_stb_i & wb_we_i & !feed_bsy & !wb_ack_q;
      wb_ack = wb_cyc_i & wb_stb_i & ((wb_we_i & !feed_bsy ) | !wb_we_i);
    end
    6'h8: wb_dat = usec_i[15:0];
    6'ha: wb_dat = usec_i[31:16];
    6'hc: wb_dat = random[15:0];
    6'he: wb_dat = random[31:16];
    6'h10: wb_dat = {1'b1, hi[4:0], wb_sel_i[1:0], 1'b0, lo[4:0], wb_sel_i[1:0]};
    6'h12: wb_dat = {sd_busy_i, dma_idle_i, 1'b0, red_led, green_led, {11{1'b0}}};
    6'h28: wb_dat = {8'hxx, 2'b00, wdogctl, tag_clk, tag_si, tag_csn, tag_so_q};
    6'h2a: wb_dat = custom;
    6'h2c: wb_dat = {evgpio2_irq, evgpio_irq, irqs_i[13:0]} & ~irq_mask;
    6'h2e: wb_dat = irq_mask;
    6'h30: wb_dat = tagmem0;
    6'h32: wb_dat = tagmem1;
    6'h34: wb_dat = {adc_a, adc_scl, adc_sda, adc_scl_i, adc_sda_i};
    6'h36: begin
      wb_dat = evgpio_dat;
      evgpio_en = 1'b1;
      wb_ack = evgpio_ack;
    end
    6'h38: begin
      wb_dat = evgpiomask_dat;
      evgpiomask_en = 1'b1;
      wb_ack = evgpiomask_ack;
    end
    6'h3a: begin
      wb_dat = evgpio2_dat;
      evgpio2_en = 1'b1;
      wb_ack = evgpio2_ack;
    end
    6'h3c: begin
      wb_dat = evgpio2mask_dat;
      evgpio2mask_en = 1'b1;
      wb_ack = evgpio2mask_ack;
    end
    6'h3e: wb_dat[3:0] = pwm;
    
    //6'h2c: wb_dat = wb_sel_i[1:0] == 2'b11 ? 16'h1616 : 16'h8888; //width test
    //default: wb_dat = 16'hxxxx;
  endcase
end

endmodule

